##############################################################################
# Copyright 2016-2018 Rigetti Computing
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################
"""Standard Gate definitions to use in a ``Program``."""

from collections.abc import Mapping, Sequence
from numbers import Real
from typing import Callable, Optional, Union, no_type_check

import numpy as np
from deprecated.sphinx import deprecated, versionadded

from pyquil.quilatom import (
    Expression,
    FormalArgument,
    Frame,
    MemoryReference,
    MemoryReferenceDesignator,
    ParameterDesignator,
    Qubit,
    QubitDesignator,
    Waveform,
    unpack_classical_reg,
    unpack_qubit,
)
from pyquil.quilbase import (
    AbstractInstruction,
    Capture,
    ClassicalAdd,
    ClassicalAnd,
    ClassicalConvert,
    ClassicalDiv,
    ClassicalEqual,
    ClassicalExchange,
    ClassicalExclusiveOr,
    ClassicalGreaterEqual,
    ClassicalGreaterThan,
    ClassicalInclusiveOr,
    ClassicalLessEqual,
    ClassicalLessThan,
    ClassicalLoad,
    ClassicalMove,
    ClassicalMul,
    ClassicalNeg,
    ClassicalNot,
    ClassicalStore,
    ClassicalSub,
    Declare,
    DelayFrames,
    DelayQubits,
    Fence,
    FenceAll,
    Gate,
    Halt,
    Measurement,
    Nop,
    Pulse,
    RawCapture,
    Reset,
    ResetQubit,
    SetFrequency,
    SetPhase,
    SetScale,
    ShiftFrequency,
    ShiftPhase,
    SwapPhases,
    Wait,
)


def unpack_reg_val_pair(
    classical_reg1: MemoryReferenceDesignator,
    classical_reg2: Union[MemoryReferenceDesignator, int, float],
) -> tuple[MemoryReference, Union[MemoryReference, int, float]]:
    """Type check/coerce arguments to constructors for binary classical operators.

    :param classical_reg1: Specifier for the classical memory address to be modified.
    :param classical_reg2: Specifier for the second argument: a classical memory address or an
        immediate value.
    :return: A pair of pyQuil objects suitable for use as operands.
    """
    left = unpack_classical_reg(classical_reg1)
    if isinstance(classical_reg2, (float, int)):
        return left, classical_reg2
    return left, unpack_classical_reg(classical_reg2)


def prepare_ternary_operands(
    classical_reg1: MemoryReferenceDesignator,
    classical_reg2: MemoryReferenceDesignator,
    classical_reg3: Union[MemoryReferenceDesignator, int, float],
) -> tuple[MemoryReference, MemoryReference, Union[MemoryReference, int, float]]:
    """Type check/coerce arguments to constructors for ternary classical operators.

    :param classical_reg1: Specifier for the classical memory address to be modified.
    :param classical_reg2: Specifier for the left operand: a classical memory address.
    :param classical_reg3: Specifier for the right operand: a classical memory address or an
        immediate value.
    :return: A triple of pyQuil objects suitable for use as operands.
    """
    if isinstance(classical_reg1, int):
        raise TypeError("Target operand of comparison must be a memory address")
    classical_reg1 = unpack_classical_reg(classical_reg1)
    if isinstance(classical_reg2, int):
        raise TypeError("Left operand of comparison must be a memory address")
    classical_reg2 = unpack_classical_reg(classical_reg2)
    if not isinstance(classical_reg3, (float, int)):
        classical_reg3 = unpack_classical_reg(classical_reg3)

    return classical_reg1, classical_reg2, classical_reg3


def I(qubit: QubitDesignator) -> Gate:  # noqa: E743 - I is not ambiguous in this case
    """Produce a single-qubit identity gate (I).

        I = [1, 0]
            [0, 1]

    This gate is a single qubit identity gate.
    Note that this gate is different that the NOP instruction as noise channels
    are typically still applied during the duration of identity gates. Identities will
    also block parallelization like any other gate.

    :param qubit: The qubit apply the gate to.
    :returns: A Gate object.
    """
    return Gate(name="I", params=[], qubits=[unpack_qubit(qubit)])


def X(qubit: QubitDesignator) -> Gate:
    """Produce the X ("NOT") gate.

        X = [[0, 1],
             [1, 0]]

    This gate is a single qubit X-gate.

    :param qubit: The qubit apply the gate to.
    :returns: A Gate object.
    """
    return Gate(name="X", params=[], qubits=[unpack_qubit(qubit)])


def Y(qubit: QubitDesignator) -> Gate:
    """Produce the Y gate.

        Y = [[0, 0 - 1j],
             [0 + 1j, 0]]

    This gate is a single qubit Y-gate.

    :param qubit: The qubit apply the gate to.
    :returns: A Gate object.
    """
    return Gate(name="Y", params=[], qubits=[unpack_qubit(qubit)])


def Z(qubit: QubitDesignator) -> Gate:
    """Produce the Z gate.

        Z = [[1,  0],
             [0, -1]]

    This gate is a single qubit Z-gate.

    :param qubit: The qubit apply the gate to.
    :returns: A Gate object.
    """
    return Gate(name="Z", params=[], qubits=[unpack_qubit(qubit)])


def H(qubit: QubitDesignator) -> Gate:
    """Produce the Hadamard gate.

        H = (1 / sqrt(2)) * [[1,  1],
                             [1, -1]]

    Produce the H instruction. This gate is a single qubit Hadamard gate.

    :param qubit: The qubit apply the gate to.
    :returns: A Gate object.
    """
    return Gate(name="H", params=[], qubits=[unpack_qubit(qubit)])


def S(qubit: QubitDesignator) -> Gate:
    """Produce the S gate.

        S = [[1, 0],
             [0, 1j]]

    This gate is a single qubit S-gate.

    :param qubit: The qubit apply the gate to.
    :returns: A Gate object.
    """
    return Gate(name="S", params=[], qubits=[unpack_qubit(qubit)])


def T(qubit: QubitDesignator) -> Gate:
    """Produce the T gate.

        T = [[1, 0],
             [0, exp(1j * pi / 4)]]

    This gate is a single qubit T-gate. It is the same as RZ(pi/4).

    :param qubit: The qubit apply the gate to.
    :returns: A Gate object.
    """
    return Gate(name="T", params=[], qubits=[unpack_qubit(qubit)])


def RX(angle: ParameterDesignator, qubit: QubitDesignator) -> Gate:
    """Produce the RX gate.

        RX(phi) = [[cos(phi / 2), -1j * sin(phi / 2)],
                   [-1j * sin(phi / 2), cos(phi / 2)]]

    This gate is a single qubit X-rotation.

    :param angle: The angle to rotate around the x-axis on the bloch sphere.
    :param qubit: The qubit apply the gate to.
    :returns: A Gate object.
    """
    return Gate(name="RX", params=[angle], qubits=[unpack_qubit(qubit)])


def RY(angle: ParameterDesignator, qubit: QubitDesignator) -> Gate:
    """Produce the RY gate.

        RY(phi) = [[cos(phi / 2), -sin(phi / 2)],
                   [sin(phi / 2), cos(phi / 2)]]

    This gate is a single qubit Y-rotation.

    :param angle: The angle to rotate around the y-axis on the bloch sphere.
    :param qubit: The qubit apply the gate to.
    :returns: A Gate object.
    """
    return Gate(name="RY", params=[angle], qubits=[unpack_qubit(qubit)])


def RZ(angle: ParameterDesignator, qubit: QubitDesignator) -> Gate:
    """Produce the RZ gate.

        RZ(phi) = [[cos(phi / 2) - 1j * sin(phi / 2), 0]
                   [0, cos(phi / 2) + 1j * sin(phi / 2)]]

    This gate is a single qubit Z-rotation.

    :param angle: The angle to rotate around the z-axis on the bloch sphere.
    :param qubit: The qubit apply the gate to.
    :returns: A Gate object.
    """
    return Gate(name="RZ", params=[angle], qubits=[unpack_qubit(qubit)])


def U(theta: ParameterDesignator, phi: ParameterDesignator, lam: ParameterDesignator, qubit: QubitDesignator) -> Gate:
    """Produce a generic single-qubit rotation.

        U(theta, phi, lam) = [[              cos(theta / 2),  -1 * exp(1j*lam) * sin(theta / 2)]
                              [exp(1j*phi) * sin(theta / 2), exp(1j*(phi+lam)) * cos(theta / 2)]]

    Single qubit rotation with 3 Euler angles.

    :param theta: The theta Euler angle.
    :param phi: The phi Euler angle.
    :param lam: The lambda Euler angle.
    :param qubit: The qubit to apply the gate to.
    :returns: A Gate object.
    """
    return Gate(name="U", params=[theta, phi, lam], qubits=[unpack_qubit(qubit)])


def PHASE(angle: ParameterDesignator, qubit: QubitDesignator) -> Gate:
    """Produce the PHASE gate.

        PHASE(phi) = [[1, 0],
                      [0, exp(1j * phi)]]

    This is the same as the RZ gate.

    :param angle: The angle to rotate around the z-axis on the bloch sphere.
    :param qubit: The qubit apply the gate to.
    :returns: A Gate object.
    """
    return Gate(name="PHASE", params=[angle], qubits=[unpack_qubit(qubit)])


def CZ(control: QubitDesignator, target: QubitDesignator) -> Gate:
    """Produce a controlled-Z gate.

        CZ = [[1, 0, 0,  0],
              [0, 1, 0,  0],
              [0, 0, 1,  0],
              [0, 0, 0, -1]]


    This gate applies to two qubit arguments to produce the controlled-Z gate instruction.

    :param control: The control qubit.
    :param target: The target qubit. The target qubit has an Z-gate applied to it if the control
        qubit is in the excited state.
    :returns: A Gate object.
    """
    return Gate(name="CZ", params=[], qubits=[unpack_qubit(q) for q in (control, target)])


def CNOT(control: QubitDesignator, target: QubitDesignator) -> Gate:
    """Produce a controlled-NOT (controlled-X) gate.

        CNOT = [[1, 0, 0, 0],
                [0, 1, 0, 0],
                [0, 0, 0, 1],
                [0, 0, 1, 0]]

    This gate applies to two qubit arguments to produce the controlled-not gate instruction.

    :param control: The control qubit.
    :param target: The target qubit. The target qubit has an X-gate applied to it if the control
        qubit is in the ``|1>`` state.
    :returns: A Gate object.
    """
    return Gate(name="CNOT", params=[], qubits=[unpack_qubit(q) for q in (control, target)])


def CCNOT(control1: QubitDesignator, control2: QubitDesignator, target: QubitDesignator) -> Gate:
    """Produce a doubly-controlled NOT gate.

        CCNOT = [[1, 0, 0, 0, 0, 0, 0, 0],
                 [0, 1, 0, 0, 0, 0, 0, 0],
                 [0, 0, 1, 0, 0, 0, 0, 0],
                 [0, 0, 0, 1, 0, 0, 0, 0],
                 [0, 0, 0, 0, 1, 0, 0, 0],
                 [0, 0, 0, 0, 0, 1, 0, 0],
                 [0, 0, 0, 0, 0, 0, 0, 1],
                 [0, 0, 0, 0, 0, 0, 1, 0]]

    This gate applies to three qubit arguments to produce the controlled-controlled-not gate
    instruction.

    :param control1: The first control qubit.
    :param control2: The second control qubit.
    :param target: The target qubit. The target qubit has an X-gate applied to it if both control
        qubits are in the excited state.
    :returns: A Gate object.
    """
    qubits = [unpack_qubit(q) for q in (control1, control2, target)]
    return Gate(name="CCNOT", params=[], qubits=qubits)


def CPHASE00(angle: ParameterDesignator, control: QubitDesignator, target: QubitDesignator) -> Gate:
    """Produce a controlled-phase gate that phases the ``|00>`` state.

        CPHASE00(phi) = diag([exp(1j * phi), 1, 1, 1])

    This gate applies to two qubit arguments to produce the variant of the controlled phase
    instruction that affects the state 00.

    :param angle: The input phase angle to apply when both qubits are in the ``|0>`` state.
    :param control: Qubit 1.
    :param target: Qubit 2.
    :returns: A Gate object.
    """
    qubits = [unpack_qubit(q) for q in (control, target)]
    return Gate(name="CPHASE00", params=[angle], qubits=qubits)


def CPHASE01(angle: ParameterDesignator, control: QubitDesignator, target: QubitDesignator) -> Gate:
    """Produce a controlled-phase gate that phases the ``|01>`` state.

        CPHASE01(phi) = diag([1.0, exp(1j * phi), 1.0, 1.0])

    This gate applies to two qubit arguments to produce the variant of the controlled phase
    instruction that affects the state 01.

    :param angle: The input phase angle to apply when q1 is in the ``|1>`` state and q2 is in
        the ``|0>`` state.
    :param control: Qubit 1.
    :param target: Qubit 2.
    :returns: A Gate object.
    """
    qubits = [unpack_qubit(q) for q in (control, target)]
    return Gate(name="CPHASE01", params=[angle], qubits=qubits)


def CPHASE10(angle: ParameterDesignator, control: QubitDesignator, target: QubitDesignator) -> Gate:
    """Produce a controlled-phase gate that phases the ``|10>`` state.

        CPHASE10(phi) = diag([1, 1, exp(1j * phi), 1])

    This gate applies to two qubit arguments to produce the variant of the controlled phase
    instruction that affects the state 10.

    :param angle: The input phase angle to apply when q2 is in the ``|1>`` state and q1 is in
        the ``|0>`` state.
    :param control: Qubit 1.
    :param target: Qubit 2.
    :returns: A Gate object.
    """
    qubits = [unpack_qubit(q) for q in (control, target)]
    return Gate(name="CPHASE10", params=[angle], qubits=qubits)


# NOTE: We don't use ParameterDesignator here because of the following Sphinx error. This error
# can be resolved by importing Expression, but then flake8 complains about an unused import:
#   Cannot resolve forward reference in type annotations of "pyquil.gates.CPHASE":
#   name 'Expression' is not defined
def CPHASE(
    angle: Union[Expression, MemoryReference, np.int_, int, float, complex],
    control: QubitDesignator,
    target: QubitDesignator,
) -> Gate:
    """Produce a controlled-phase instruction.

        CPHASE(phi) = diag([1, 1, 1, exp(1j * phi)])

    This gate applies to two qubit arguments to produce the variant of the controlled phase
    instruction that affects the state 11.

    Compare with the ``CPHASExx`` variants. This variant is the most common and does
    not have a suffix, although you can think of it as ``CPHASE11``.

    :param angle: The input phase angle to apply when both qubits are in the ``|1>`` state.
    :param control: Qubit 1.
    :param target: Qubit 2.
    :returns: A Gate object.
    """
    qubits = [unpack_qubit(q) for q in (control, target)]
    return Gate(name="CPHASE", params=[angle], qubits=qubits)


def SWAP(q1: QubitDesignator, q2: QubitDesignator) -> Gate:
    """Produce a SWAP gate which swaps the state of two qubits.

        SWAP = [[1, 0, 0, 0],
                [0, 0, 1, 0],
                [0, 1, 0, 0],
                [0, 0, 0, 1]]


    :param q1: Qubit 1.
    :param q2: Qubit 2.
    :returns: A Gate object.
    """
    return Gate(name="SWAP", params=[], qubits=[unpack_qubit(q) for q in (q1, q2)])


def CSWAP(control: QubitDesignator, target_1: QubitDesignator, target_2: QubitDesignator) -> Gate:
    """Produce a controlled-SWAP gate. This gate conditionally swaps the state of two qubits.

        CSWAP = [[1, 0, 0, 0, 0, 0, 0, 0],
                 [0, 1, 0, 0, 0, 0, 0, 0],
                 [0, 0, 1, 0, 0, 0, 0, 0],
                 [0, 0, 0, 1, 0, 0, 0, 0],
                 [0, 0, 0, 0, 1, 0, 0, 0],
                 [0, 0, 0, 0, 0, 0, 1, 0],
                 [0, 0, 0, 0, 0, 1, 0, 0],
                 [0, 0, 0, 0, 0, 0, 0, 1]]


    :param control: The control qubit.
    :param target_1: The first target qubit.
    :param target_2: The second target qubit. The two target states are swapped if the control is
        in the ``|1>`` state.
    """
    qubits = [unpack_qubit(q) for q in (control, target_1, target_2)]
    return Gate(name="CSWAP", params=[], qubits=qubits)


def ISWAP(q1: QubitDesignator, q2: QubitDesignator) -> Gate:
    """Produce an ISWAP gate.

        ISWAP = [[1, 0,  0,  0],
                 [0, 0,  1j, 0],
                 [0, 1j, 0,  0],
                 [0, 0,  0,  1]]

    This gate swaps the state of two qubits, applying a -i phase to q1 when it
    is in the 1 state and a -i phase to q2 when it is in the 0 state.

    :param q1: Qubit 1.
    :param q2: Qubit 2.
    :returns: A Gate object.
    """
    return Gate(name="ISWAP", params=[], qubits=[unpack_qubit(q) for q in (q1, q2)])


def PSWAP(angle: ParameterDesignator, q1: QubitDesignator, q2: QubitDesignator) -> Gate:
    """Produce a parameterized SWAP gate.

        PSWAP(phi) = [[1, 0,             0,             0],
                      [0, 0,             exp(1j * phi), 0],
                      [0, exp(1j * phi), 0,             0],
                      [0, 0,             0,             1]]


    :param angle: The angle of the phase to apply to the swapped states. This phase is applied to
        q1 when it is in the 1 state and to q2 when it is in the 0 state.
    :param q1: Qubit 1.
    :param q2: Qubit 2.
    :returns: A Gate object.
    """
    return Gate(name="PSWAP", params=[angle], qubits=[unpack_qubit(q) for q in (q1, q2)])


def XY(angle: ParameterDesignator, q1: QubitDesignator, q2: QubitDesignator) -> Gate:
    """Produce a parameterized ISWAP gate.

        XY(phi) = [[1,               0,               0, 0],
                   [0,      cos(phi/2), 1j * sin(phi/2), 0],
                   [0, 1j * sin(phi/2),      cos(phi/2), 0],
                   [0,               0,               0, 1]]

    :param angle: The angle of the rotation to apply to the population 1 subspace.
    :param q1: Qubit 1.
    :param q2: Qubit 2.
    :returns: A Gate object.
    """
    return Gate(name="XY", params=[angle], qubits=[unpack_qubit(q) for q in (q1, q2)])


def SQISW(q1: QubitDesignator, q2: QubitDesignator) -> Gate:
    """Produce a SQISW gate.

        SQiSW = [[1,               0,               0, 0],
                 [0,     1 / sqrt(2),    1j / sqrt(2), 0],
                 [0,    1j / sqrt(2),     1 / sqrt(2), 0],
                 [0,               0,               0, 1]]

    :param q1: Qubit 1.
    :param q2: Qubit 2.
    :returns: A Gate object.
    """
    return Gate(name="SQISW", params=[], qubits=[unpack_qubit(q) for q in (q1, q2)])


def FSIM(theta: ParameterDesignator, phi: ParameterDesignator, q1: QubitDesignator, q2: QubitDesignator) -> Gate:
    """Produce an fsim (Fermionic simulation) gate.

        FSIM(theta, phi) = [[1,                 0,                 0,           0],
                            [0,      cos(theta/2), 1j * sin(theta/2),           0],
                            [0, 1j * sin(theta/2),      cos(theta/2),           0],
                            [0,                 0,                 0, exp(1j*phi)]]

    :param theta: The angle for the XX + YY rotation.
    :param phi: The angle for the ZZ rotation.
    :param q1: Qubit 1.
    :param q2: Qubit 2.
    :returns: A Gate object.
    """
    return Gate(name="FSIM", params=[theta, phi], qubits=[unpack_qubit(q) for q in (q1, q2)])


def PHASEDFSIM(
    theta: ParameterDesignator,
    zeta: ParameterDesignator,
    chi: ParameterDesignator,
    gamma: ParameterDesignator,
    phi: ParameterDesignator,
    q1: QubitDesignator,
    q2: QubitDesignator,
) -> Gate:
    """Produce an phasedfsim (Fermionic simulation) gate.

        PHASEDFSIM(theta, zeta, chi, gamma, phi) = [
            [1, 0, 0, 0],
            [0, exp(-1j*(gamma+zeta)) * cos(theta/2), 1j* exp(-1j*(gamma-chi)) * sin(theta/2), 0],
            [0, 1j* exp(-1j*(gamma+chi)) * sin(theta/2),     exp(-1j*(gamma-zeta)) * cos(theta/2), 0],
            [0, 0, 0, exp(1j*phi - 2j*gamma)]]

    :param theta: The angle for the XX + YY rotation.
    :param zeta: Zeta phase.
    :param chi: Chi phase.
    :param gamma: Gamma phase.
    :param phi: The angle for the ZZ rotation.
    :param q1: Qubit 1.
    :param q2: Qubit 2.
    :returns: A Gate object.
    """
    return Gate(name="PHASEDFSIM", params=[theta, zeta, chi, gamma, phi], qubits=[unpack_qubit(q) for q in (q1, q2)])


def RZZ(phi: ParameterDesignator, q1: QubitDesignator, q2: QubitDesignator) -> Gate:
    """Produce a RZZ gate.

        RZZ(phi) = [[ exp(-1j*phi/2),             0,             0,              0],
                    [              0, exp(1j*phi/2),             0,              0],
                    [              0,             0, exp(1j*phi/2),              0],
                    [              0,             0,             0, exp(-1j*phi/2)]]

    :param phi: The angle for the ZZ rotation.
    :param q1: Qubit 1.
    :param q2: Qubit 2.
    :returns: A Gate object.
    """
    return Gate(name="RZZ", params=[phi], qubits=[unpack_qubit(q) for q in (q1, q2)])


def RXX(phi: ParameterDesignator, q1: QubitDesignator, q2: QubitDesignator) -> Gate:
    """Produce a RXX gate.

        RXX(phi) = [[     cos(phi/2),              0,              0, -1j*sin(phi/2)],
                    [              0,     cos(phi/2), -1j*sin(phi/2),              0],
                    [              0, -1j*sin(phi/2),     cos(phi/2),              0],
                    [ -1j*sin(phi/2),              0,              0,     cos(phi/2)]]

    :param phi: The angle for the XX rotation.
    :param q1: Qubit 1.
    :param q2: Qubit 2.
    :returns: A Gate object.
    """
    return Gate(name="RXX", params=[phi], qubits=[unpack_qubit(q) for q in (q1, q2)])


def RYY(phi: ParameterDesignator, q1: QubitDesignator, q2: QubitDesignator) -> Gate:
    """Produce a RYY gate.

        RYY(phi) = [[    cos(phi/2),              0,              0, 1j*sin(phi/2)],
                    [             0,     cos(phi/2), -1j*sin(phi/2),             0],
                    [             0, -1j*sin(phi/2),     cos(phi/2),             0],
                    [ 1j*sin(phi/2),              0,              0,    cos(phi/2)]]

    :param phi: The angle for the YY rotation.
    :param q1: Qubit 1.
    :param q2: Qubit 2.
    :returns: A Gate object.
    """
    return Gate(name="RYY", params=[phi], qubits=[unpack_qubit(q) for q in (q1, q2)])


WAIT = Wait()
"""
This instruction tells the quantum computation to halt. Typically these is used while classical
memory is being manipulated by a CPU in a hybrid classical/quantum algorithm.

:returns: A Wait object.
"""


def RESET(qubit_index: Optional[QubitDesignator] = None) -> Union[Reset, ResetQubit]:
    """Reset all qubits or just one specific qubit.

    :param qubit_index: The qubit to reset.
        This can be a qubit's index, a Qubit, or a QubitPlaceholder.
        If None, reset all qubits.
    :returns: A Reset or ResetQubit Quil AST expression corresponding to a global or targeted
        reset, respectively.
    """
    if qubit_index is not None:
        return ResetQubit(unpack_qubit(qubit_index))
    else:
        return Reset()


NOP = Nop()
"""
This instruction applies no operation at that timestep. Typically these are ignored in error-models.

:returns: A Nop object.
"""

HALT = Halt()
"""
This instruction ends the program.

:returns: A Halt object.
"""


def DECLARE(
    name: str,
    memory_type: str = "BIT",
    memory_size: int = 1,
    shared_region: Optional[str] = None,
    offsets: Optional[Sequence[tuple[int, str]]] = None,
) -> Declare:
    return Declare(
        name=name,
        memory_type=memory_type,
        memory_size=memory_size,
        shared_region=shared_region,
        offsets=offsets,
    )


def MEASURE(qubit: QubitDesignator, classical_reg: Optional[MemoryReferenceDesignator]) -> Measurement:
    """Produce a MEASURE instruction.

    :param qubit: The qubit to measure.
    :param classical_reg: The classical register to measure into, or None.
    :return: A Measurement instance.
    """
    qubit = unpack_qubit(qubit)
    if classical_reg is None:
        address = None
    else:
        address = unpack_classical_reg(classical_reg)
    return Measurement(qubit, address)


def NEG(classical_reg: MemoryReferenceDesignator) -> ClassicalNeg:
    """Produce a NEG instruction.

    :param classical_reg: A classical memory address to modify.
    :return: A ClassicalNeg instance.
    """
    return ClassicalNeg(unpack_classical_reg(classical_reg))


def NOT(classical_reg: MemoryReferenceDesignator) -> ClassicalNot:
    """Produce a NOT instruction.

    :param classical_reg: A classical register to modify.
    :return: A ClassicalNot instance.
    """
    return ClassicalNot(unpack_classical_reg(classical_reg))


def AND(
    classical_reg1: MemoryReferenceDesignator, classical_reg2: Union[MemoryReferenceDesignator, int]
) -> ClassicalAnd:
    """Produce an AND instruction.

    NOTE: The order of operands was reversed in pyQuil <=1.9 .

    :param classical_reg1: The first classical register, which gets modified.
    :param classical_reg2: The second classical register or immediate value.
    :return: A ClassicalAnd instance.
    """
    left, right = unpack_reg_val_pair(classical_reg1, classical_reg2)
    if not isinstance(right, (MemoryReference, int)):
        raise TypeError("Right operand of AND must be a memory address or integer.")
    return ClassicalAnd(left, right)


def IOR(
    classical_reg1: MemoryReferenceDesignator, classical_reg2: Union[MemoryReferenceDesignator, int]
) -> ClassicalInclusiveOr:
    """Produce an inclusive OR instruction.

    :param classical_reg1: The first classical register, which gets modified.
    :param classical_reg2: The second classical register or immediate value.
    :return: A ClassicalInclusiveOr instance.
    """
    left, right = unpack_reg_val_pair(classical_reg1, classical_reg2)
    if not isinstance(right, (MemoryReference, int)):
        raise TypeError("Right operand of AND must be a memory address or integer.")
    return ClassicalInclusiveOr(left, right)


def XOR(
    classical_reg1: MemoryReferenceDesignator, classical_reg2: Union[MemoryReferenceDesignator, int]
) -> ClassicalExclusiveOr:
    """Produce an exclusive OR instruction.

    :param classical_reg1: The first classical register, which gets modified.
    :param classical_reg2: The second classical register or immediate value.
    :return: A ClassicalExclusiveOr instance.
    """
    left, right = unpack_reg_val_pair(classical_reg1, classical_reg2)
    if not isinstance(right, (MemoryReference, int)):
        raise TypeError("Right operand of AND must be a memory address or integer.")
    return ClassicalExclusiveOr(left, right)


def MOVE(
    classical_reg1: MemoryReferenceDesignator,
    classical_reg2: Union[MemoryReferenceDesignator, int, float],
) -> ClassicalMove:
    """Produce a MOVE instruction.

    :param classical_reg1: The first classical register, which gets modified.
    :param classical_reg2: The second classical register or immediate value.
    :return: A ClassicalMove instance.
    """
    left, right = unpack_reg_val_pair(classical_reg1, classical_reg2)
    return ClassicalMove(left, right)


def EXCHANGE(classical_reg1: MemoryReferenceDesignator, classical_reg2: MemoryReferenceDesignator) -> ClassicalExchange:
    """Produce an EXCHANGE instruction.

    :param classical_reg1: The first classical register, which gets modified.
    :param classical_reg2: The second classical register, which gets modified.
    :return: A ClassicalExchange instance.
    """
    left = unpack_classical_reg(classical_reg1)
    right = unpack_classical_reg(classical_reg2)
    return ClassicalExchange(left, right)


def LOAD(
    target_reg: MemoryReferenceDesignator, region_name: str, offset_reg: MemoryReferenceDesignator
) -> ClassicalLoad:
    """Produce a LOAD instruction.

    :param target_reg: LOAD storage target.
    :param region_name: Named region of memory to load from.
    :param offset_reg: Offset into region of memory to load from. Must be a MemoryReference.
    :return: A ClassicalLoad instance.
    """
    return ClassicalLoad(unpack_classical_reg(target_reg), region_name, unpack_classical_reg(offset_reg))


def STORE(
    region_name: str,
    offset_reg: MemoryReferenceDesignator,
    source: Union[MemoryReferenceDesignator, int, float],
) -> ClassicalStore:
    """Produce a STORE instruction.

    :param region_name: Named region of memory to store to.
    :param offset_reg: Offset into memory region. Must be a MemoryReference.
    :param source: Source data. Can be either a MemoryReference or a constant.
    :return: A ClassicalStore instance.
    """
    if not isinstance(source, int) and not isinstance(source, float):
        source = unpack_classical_reg(source)
    return ClassicalStore(region_name, unpack_classical_reg(offset_reg), source)


def CONVERT(classical_reg1: MemoryReferenceDesignator, classical_reg2: MemoryReferenceDesignator) -> ClassicalConvert:
    """Produce a CONVERT instruction.

    :param classical_reg1: MemoryReference to store to.
    :param classical_reg2: MemoryReference to read from.
    :return: A ClassicalConvert instance.
    """
    return ClassicalConvert(unpack_classical_reg(classical_reg1), unpack_classical_reg(classical_reg2))


def ADD(classical_reg: MemoryReferenceDesignator, right: Union[MemoryReferenceDesignator, int, float]) -> ClassicalAdd:
    """Produce an ADD instruction.

    :param classical_reg: Left operand for the arithmetic operation. Also serves as the store
        target.
    :param right: Right operand for the arithmetic operation.
    :return: A ClassicalAdd instance.
    """
    left, right = unpack_reg_val_pair(classical_reg, right)
    return ClassicalAdd(left, right)


def SUB(classical_reg: MemoryReferenceDesignator, right: Union[MemoryReferenceDesignator, int, float]) -> ClassicalSub:
    """Produce a SUB instruction.

    :param classical_reg: Left operand for the arithmetic operation. Also serves as the store
        target.
    :param right: Right operand for the arithmetic operation.
    :return: A ClassicalSub instance.
    """
    left, right = unpack_reg_val_pair(classical_reg, right)
    return ClassicalSub(left, right)


def MUL(classical_reg: MemoryReferenceDesignator, right: Union[MemoryReferenceDesignator, int, float]) -> ClassicalMul:
    """Produce a MUL instruction.

    :param classical_reg: Left operand for the arithmetic operation. Also serves as the store
        target.
    :param right: Right operand for the arithmetic operation.
    :return: A ClassicalMul instance.
    """
    left, right = unpack_reg_val_pair(classical_reg, right)
    return ClassicalMul(left, right)


def DIV(classical_reg: MemoryReferenceDesignator, right: Union[MemoryReferenceDesignator, int, float]) -> ClassicalDiv:
    """Produce an DIV instruction.

    :param classical_reg: Left operand for the arithmetic operation. Also serves as the store
        target.
    :param right: Right operand for the arithmetic operation.
    :return: A ClassicalDiv instance.
    """
    left, right = unpack_reg_val_pair(classical_reg, right)
    return ClassicalDiv(left, right)


def EQ(
    classical_reg1: MemoryReferenceDesignator,
    classical_reg2: MemoryReferenceDesignator,
    classical_reg3: Union[MemoryReferenceDesignator, int, float],
) -> ClassicalEqual:
    """Produce an EQ instruction.

    :param classical_reg1: Memory address to which to store the comparison result.
    :param classical_reg2: Left comparison operand.
    :param classical_reg3: Right comparison operand.
    :return: A ClassicalEqual instance.
    """
    classical_reg1, classical_reg2, classical_reg3 = prepare_ternary_operands(
        classical_reg1, classical_reg2, classical_reg3
    )

    return ClassicalEqual(classical_reg1, classical_reg2, classical_reg3)


def LT(
    classical_reg1: MemoryReferenceDesignator,
    classical_reg2: MemoryReferenceDesignator,
    classical_reg3: Union[MemoryReferenceDesignator, int, float],
) -> ClassicalLessThan:
    """Produce an LT instruction.

    :param classical_reg1: Memory address to which to store the comparison result.
    :param classical_reg2: Left comparison operand.
    :param classical_reg3: Right comparison operand.
    :return: A ClassicalLessThan instance.
    """
    classical_reg1, classical_reg2, classical_reg3 = prepare_ternary_operands(
        classical_reg1, classical_reg2, classical_reg3
    )
    return ClassicalLessThan(classical_reg1, classical_reg2, classical_reg3)


def LE(
    classical_reg1: MemoryReferenceDesignator,
    classical_reg2: MemoryReferenceDesignator,
    classical_reg3: Union[MemoryReferenceDesignator, int, float],
) -> ClassicalLessEqual:
    """Produce an LE instruction.

    :param classical_reg1: Memory address to which to store the comparison result.
    :param classical_reg2: Left comparison operand.
    :param classical_reg3: Right comparison operand.
    :return: A ClassicalLessEqual instance.
    """
    classical_reg1, classical_reg2, classical_reg3 = prepare_ternary_operands(
        classical_reg1, classical_reg2, classical_reg3
    )
    return ClassicalLessEqual(classical_reg1, classical_reg2, classical_reg3)


def GT(
    classical_reg1: MemoryReferenceDesignator,
    classical_reg2: MemoryReferenceDesignator,
    classical_reg3: Union[MemoryReferenceDesignator, int, float],
) -> ClassicalGreaterThan:
    """Produce an GT instruction.

    :param classical_reg1: Memory address to which to store the comparison result.
    :param classical_reg2: Left comparison operand.
    :param classical_reg3: Right comparison operand.
    :return: A ClassicalGreaterThan instance.
    """
    classical_reg1, classical_reg2, classical_reg3 = prepare_ternary_operands(
        classical_reg1, classical_reg2, classical_reg3
    )
    return ClassicalGreaterThan(classical_reg1, classical_reg2, classical_reg3)


def GE(
    classical_reg1: MemoryReferenceDesignator,
    classical_reg2: MemoryReferenceDesignator,
    classical_reg3: Union[MemoryReferenceDesignator, int, float],
) -> ClassicalGreaterEqual:
    """Produce an GE instruction.

    :param classical_reg1: Memory address to which to store the comparison result.
    :param classical_reg2: Left comparison operand.
    :param classical_reg3: Right comparison operand.
    :return: A ClassicalGreaterEqual instance.
    """
    classical_reg1, classical_reg2, classical_reg3 = prepare_ternary_operands(
        classical_reg1, classical_reg2, classical_reg3
    )
    return ClassicalGreaterEqual(classical_reg1, classical_reg2, classical_reg3)


def PULSE(frame: Frame, waveform: Waveform, nonblocking: bool = False) -> Pulse:
    """Produce a PULSE instruction.

    :param frame: The frame on which to apply the pulse.
    :param waveform: The pulse waveform.
    :param nonblocking: A flag indicating whether the pulse is NONBLOCKING.
    :return: A Pulse instance.
    """
    return Pulse(frame, waveform, nonblocking)


def SET_FREQUENCY(frame: Frame, freq: ParameterDesignator) -> SetFrequency:
    """Produce a SET-FREQUENCY instruction.

    :param frame: The frame on which to set the frequency.
    :param freq: The frequency value, in Hz.
    :returns: A SetFrequency instance.
    """
    return SetFrequency(frame, freq)


def SHIFT_FREQUENCY(frame: Frame, freq: ParameterDesignator) -> ShiftFrequency:
    """Produce a SHIFT-FREQUENCY instruction.

    :param frame: The frame on which to shift the frequency.
    :param freq: The value, in Hz, to add to the existing frequency.
    :returns: A ShiftFrequency instance.
    """
    return ShiftFrequency(frame, freq)


def SET_PHASE(frame: Frame, phase: ParameterDesignator) -> SetPhase:
    """Produce a SET-PHASE instruction.

    :param frame: The frame on which to set the phase.
    :param phase: The new phase value, in radians.
    :returns: A SetPhase instance.
    """
    return SetPhase(frame, phase)


def SHIFT_PHASE(frame: Frame, phase: ParameterDesignator) -> ShiftPhase:
    """Produce a SHIFT-PHASE instruction.

    :param frame: The frame on which to shift the phase.
    :param phase: The value, in radians, to add to the existing phase.
    :returns: A ShiftPhase instance.
    """
    return ShiftPhase(frame, phase)


@versionadded(version="3.5.1", reason="The correct instruction is SWAP-PHASES, not SWAP-PHASE")
def SWAP_PHASES(frameA: Frame, frameB: Frame) -> SwapPhases:
    """Produce a SWAP-PHASES instruction.

    :param frameA: A frame.
    :param frameB: A frame.
    :returns: A SwapPhases instance.
    """
    return SwapPhases(frameA, frameB)


@deprecated(version="3.5.1", reason="The correct instruction is SWAP-PHASES, not SWAP-PHASE")
def SWAP_PHASE(frameA: Frame, frameB: Frame) -> SwapPhases:
    """Alias of :func:`SWAP_PHASES`."""
    return SWAP_PHASES(frameA, frameB)


def SET_SCALE(frame: Frame, scale: ParameterDesignator) -> SetScale:
    """Produce a SET-SCALE instruction.

    :param frame: The frame on which to set the scale.
    :param scale: The scaling factor.
    :returns: A SetScale instance.
    """
    return SetScale(frame, scale)


def CAPTURE(
    frame: Frame,
    kernel: Waveform,
    memory_region: MemoryReferenceDesignator,
    nonblocking: bool = False,
) -> Capture:
    """Produce a CAPTURE instruction.

    :param frame: The frame on which to capture an IQ value.
    :param kernel: The integrating kernel for the capture.
    :param memory_region: The classical memory region to store the resulting IQ value.
    :param nonblocking: A flag indicating whether the capture is NONBLOCKING.
    :returns: A Capture instance.
    """
    memory_region = unpack_classical_reg(memory_region)
    return Capture(frame, kernel, memory_region, nonblocking)


def RAW_CAPTURE(
    frame: Frame,
    duration: float,
    memory_region: MemoryReferenceDesignator,
    nonblocking: bool = False,
) -> RawCapture:
    """Produce a RAW-CAPTURE instruction.

    :param frame: The frame on which to capture raw values.
    :param duration: The duration of the capture, in seconds.
    :param memory_region: The classical memory region to store the resulting raw values.
    :param nonblocking: A flag indicating whether the capture is NONBLOCKING.
    :returns: A RawCapture instance.
    """
    memory_region = unpack_classical_reg(memory_region)
    return RawCapture(frame, duration, memory_region, nonblocking)


# Mypy doesn't support a complex type hint here on args. Particularly,
# you can't tell Mypy that args should always begin with a int, end
# with a float, and everything in between should be of a particular
# type T.
@no_type_check
def DELAY(*args) -> Union[DelayFrames, DelayQubits]:
    """Produce a DELAY instruction.

    Note: There are two variants of DELAY. One applies to specific frames on some
    qubit, e.g. `DELAY 0 "rf" "ff" 1.0` delays the `"rf"` and `"ff"` frames on 0.
    It is also possible to delay all frames on some qubits, e.g. `DELAY 0 1 2 1.0`.

    :param args: A list of delay targets, ending with a duration.
    :returns: A DelayFrames or DelayQubits instance.
    """
    if len(args) < 2:
        raise ValueError(
            "Expected DELAY(t1,...,tn, duration). In particular, there "
            "must be at least one target, as well as a duration."
        )
    targets, duration = args[:-1], args[-1]
    if not isinstance(duration, (Expression, Real)):
        raise TypeError("The last argument of DELAY must be a real or parametric duration.")

    if all(isinstance(t, Frame) for t in targets):
        return DelayFrames(targets, duration)
    elif all(isinstance(t, (int, Qubit, FormalArgument)) for t in targets):
        targets = [Qubit(t) if isinstance(t, int) else t for t in targets]
        return DelayQubits(targets, duration)
    else:
        raise TypeError(
            "DELAY targets must be either (i) a list of frames, or "
            "(ii) a list of qubits / formal arguments. "
            f"Got {args}."
        )


def FENCE(*qubits: Union[int, Qubit, FormalArgument]) -> Union[FenceAll, Fence]:
    """Produce a FENCE instruction.

    Note: If no qubits are specified, then this is interpreted as a global FENCE.

    :params qubits: A list of qubits or formal arguments.
    :returns: A Fence or FenceAll instance.
    """
    if qubits:
        return Fence([Qubit(t) if isinstance(t, int) else t for t in qubits])
    else:
        return FenceAll()


QUANTUM_GATES: Mapping[str, Callable[..., Gate]] = {
    "I": I,
    "X": X,
    "Y": Y,
    "Z": Z,
    "H": H,
    "S": S,
    "T": T,
    "PHASE": PHASE,
    "RX": RX,
    "RY": RY,
    "RZ": RZ,
    "CZ": CZ,
    "CNOT": CNOT,
    "CCNOT": CCNOT,
    "CPHASE00": CPHASE00,
    "CPHASE01": CPHASE01,
    "CPHASE10": CPHASE10,
    "CPHASE": CPHASE,
    "SWAP": SWAP,
    "CSWAP": CSWAP,
    "ISWAP": ISWAP,
    "PSWAP": PSWAP,
    "XY": XY,
}
"""
Dictionary of quantum gate functions keyed by gate names.
"""

STANDARD_GATES = QUANTUM_GATES
"""
Alias for the above dictionary of quantum gates.
"""


QUILT_INSTRUCTIONS: Mapping[str, Callable[..., AbstractInstruction]] = {
    "PULSE": PULSE,
    "SET-FREQUENCY": SET_FREQUENCY,
    "SHIFT-FREQUENCY": SHIFT_FREQUENCY,
    "SET-PHASE": SET_PHASE,
    "SHIFT-PHASE": SHIFT_PHASE,
    "SWAP-PHASE": SWAP_PHASES,
    "SWAP-PHASES": SWAP_PHASES,
    "SET-SCALE": SET_SCALE,
    "CAPTURE": CAPTURE,
    "RAW-CAPTURE": RAW_CAPTURE,
    "DELAY": DELAY,
    "FENCE": FENCE,
}
"""
Dictionary of Quil-T AST construction functions keyed by instruction name.
"""

STANDARD_INSTRUCTIONS: Mapping[str, Union[AbstractInstruction, Callable[..., AbstractInstruction]]] = {
    "WAIT": WAIT,
    "RESET": RESET,
    "DECLARE": DECLARE,
    "NOP": NOP,
    "HALT": HALT,
    "MEASURE": MEASURE,
    "NOT": NOT,
    "AND": AND,
    "MOVE": MOVE,
    "EXCHANGE": EXCHANGE,
    "IOR": IOR,
    "XOR": XOR,
    "NEG": NEG,
    "ADD": ADD,
    "SUB": SUB,
    "MUL": MUL,
    "DIV": DIV,
    "EQ": EQ,
    "GT": GT,
    "GE": GE,
    "LE": LE,
    "LT": LT,
    "LOAD": LOAD,
    "STORE": STORE,
    "CONVERT": CONVERT,
}
"""
Dictionary of standard instruction functions keyed by instruction names.
"""

__all__ = (
    list(QUANTUM_GATES.keys())
    + list(fn.__name__ for fn in QUILT_INSTRUCTIONS.values())
    + list(STANDARD_INSTRUCTIONS.keys())
    + ["Gate", "QUANTUM_GATES", "STANDARD_GATES", "QUILT_INSTRUCTIONS", "STANDARD_INSTRUCTIONS"]
)



"""A pure Python implementation of the Quantum Virtual Machine (QVM)."""

##############################################################################
# Copyright 2018 Rigetti Computing
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################
import logging
from abc import ABC, abstractmethod
from collections.abc import Iterable, Sequence
from typing import Any, Optional, Union

import numpy as np
from numpy.random.mtrand import RandomState
from qcs_sdk import ExecutionData, RegisterData, ResultData
from qcs_sdk.qvm import QVMResultData

from pyquil.api import QAM, MemoryMap, QAMExecutionResult, QuantumExecutable
from pyquil.paulis import PauliSum, PauliTerm
from pyquil.quil import Program
from pyquil.quilatom import Label, LabelPlaceholder, MemoryReference
from pyquil.quilbase import (
    ArithmeticBinaryOp,
    ClassicalAdd,
    ClassicalAnd,
    ClassicalDiv,
    ClassicalExchange,
    ClassicalExclusiveOr,
    ClassicalInclusiveOr,
    ClassicalMove,
    ClassicalMul,
    ClassicalNeg,
    ClassicalNot,
    ClassicalSub,
    Declare,
    DefGate,
    DefGateByPaulis,
    DefPermutationGate,
    Gate,
    Halt,
    Jump,
    JumpTarget,
    JumpUnless,
    JumpWhen,
    LogicalBinaryOp,
    Measurement,
    Nop,
    Pragma,
    Reset,
    ResetQubit,
    UnaryClassicalInstruction,
    Wait,
)

log = logging.getLogger(__name__)

QUIL_TO_NUMPY_DTYPE = {"INT": np.int_, "REAL": np.float_, "BIT": np.int8, "OCTET": np.uint8}


class AbstractQuantumSimulator(ABC):
    """An abstract interface for a quantum simulator."""

    @abstractmethod
    def __init__(self, n_qubits: int, rs: Optional[RandomState]):
        """Initialize.

        :param n_qubits: Number of qubits to simulate.
        :param rs: a RandomState (shared with the owning :py:class:`PyQVM`) for
            doing anything stochastic.
        """

    @abstractmethod
    def do_gate(self, gate: Gate) -> "AbstractQuantumSimulator":
        """Perform a gate.

        :return: ``self`` to support method chaining.
        """

    @abstractmethod
    def do_gate_matrix(self, matrix: np.ndarray, qubits: Sequence[int]) -> "AbstractQuantumSimulator":
        """Apply an arbitrary unitary; not necessarily a named gate.

        :param matrix: The unitary matrix to apply. No checks are done
        :param qubits: A list of qubits to apply the unitary to.
        :return: ``self`` to support method chaining.
        """

    def do_program(self, program: Program) -> "AbstractQuantumSimulator":
        """Perform a sequence of gates contained within a program.

        :param program: The program
        :return: self
        """
        for gate in program:
            if not isinstance(gate, Gate):
                raise ValueError("Can only compute the simulate a program composed of `Gate`s")
            self.do_gate(gate)
        return self

    @abstractmethod
    def do_measurement(self, qubit: int) -> int:
        """Measure a qubit and collapse the wavefunction.

        :return: The measurement result. A 1 or a 0.
        """

    @abstractmethod
    def expectation(self, operator: Union[PauliTerm, PauliSum]) -> complex:
        """Compute the expectation of an operator.

        :param operator: The operator
        :return: The operator's expectation value
        """

    @abstractmethod
    def reset(self) -> "AbstractQuantumSimulator":
        """Reset the wavefunction to the ``|000...00>`` state.

        :return: ``self`` to support method chaining.
        """

    @abstractmethod
    def sample_bitstrings(self, n_samples: int) -> np.ndarray:
        """Sample bitstrings from the current state.

        :param n_samples: The number of bitstrings to sample
        :return: A numpy array of shape (n_samples, n_qubits)
        """

    @abstractmethod
    def do_post_gate_noise(self, noise_type: str, noise_prob: float, qubits: list[int]) -> "AbstractQuantumSimulator":
        """Apply noise that happens after each gate application.

        WARNING! This is experimental and the signature of this interface will likely change.

        :param noise_type: The name of the noise type
        :param noise_prob: The probability of that noise happening
        :param qubits: Apply noise to these qubits.
        :return: ``self`` to support method chaining
        """


class PyQVM(QAM["PyQVM"]):
    """A pure python implementation of the Quantum Virtual Machine."""

    def __init__(
        self,
        n_qubits: int,
        quantum_simulator_type: Optional[type[AbstractQuantumSimulator]] = None,
        seed: Optional[int] = None,
        post_gate_noise_probabilities: Optional[dict[str, float]] = None,
    ):
        """PyQuil's built-in Quil virtual machine.

        This class implements common control flow and plumbing and dispatches the "actual" work to
        quantum simulators like ReferenceWavefunctionSimulator, ReferenceDensitySimulator,
        and NumpyWavefunctionSimulator

        :param n_qubits: The number of qubits. Typically this results in the allocation of a large
            ndarray, so be judicious.
        :param quantum_simulator_type: A class that can be instantiated to handle the quantum
            aspects of this QVM. If not specified, the default will be either
            NumpyWavefunctionSimulator (no noise) or ReferenceDensitySimulator (noise)
        :param post_gate_noise_probabilities: A specification of noise model given by
            probabilities of certain types of noise. The dictionary keys are from "relaxation",
            "dephasing", "depolarizing", "phase_flip", "bit_flip", and "bitphase_flip".
            WARNING: experimental. This interface will likely change.
        :param seed: An optional random seed for performing stochastic aspects of the QVM.
        """
        if quantum_simulator_type is None:
            if post_gate_noise_probabilities is None:
                from pyquil.simulation._numpy import NumpyWavefunctionSimulator

                quantum_simulator_type = NumpyWavefunctionSimulator
            else:
                from pyquil.simulation._reference import ReferenceDensitySimulator

                log.info("Using ReferenceDensitySimulator as the backend for PyQVM")
                quantum_simulator_type = ReferenceDensitySimulator

        self.n_qubits = n_qubits
        self.ram: dict[str, list[Union[float, int]]] = {}

        if post_gate_noise_probabilities is None:
            post_gate_noise_probabilities = {}
        self.post_gate_noise_probabilities = post_gate_noise_probabilities

        self.program: Optional[Program] = None
        self.program_counter: int = 0
        self.defined_gates: dict[str, np.ndarray] = dict()

        # private implementation details
        self._qubit_to_ram: Optional[dict[int, int]] = None
        self._ro_size: Optional[int] = None
        self._memory_results = {}  # type: ignore

        self.rs = np.random.RandomState(seed=seed)
        self.wf_simulator = quantum_simulator_type(n_qubits=n_qubits, rs=self.rs)
        self._last_measure_program_loc = None

    def _extract_defined_gates(self) -> None:
        self.defined_gates = dict()
        if self.program is None:
            raise RuntimeError("No program loaded into PyQVM.")
        for dg in self.program.defined_gates:
            if dg.parameters is not None and len(dg.parameters) > 0:
                raise NotImplementedError("PyQVM does not support parameterized DEFGATEs")
            if isinstance(dg, DefPermutationGate) or isinstance(dg, DefGateByPaulis):
                raise NotImplementedError("PyQVM does not support DEFGATE ... AS MATRIX | PAULI-SUM.")
            self.defined_gates[dg.name] = dg.matrix

    def execute_with_memory_map_batch(
        self, executable: QuantumExecutable, memory_maps: Iterable[MemoryMap], **__: Any
    ) -> list["PyQVM"]:
        """Operation is not supported by PyQVM as the state of the instance is reset at the start of each execution."""
        raise NotImplementedError(
            "PyQVM does not support batch execution as the state of the instance is reset at the start of each execute."
        )

    def execute(self, executable: QuantumExecutable, memory_map: Optional[MemoryMap] = None, **__: Any) -> "PyQVM":
        """Execute a program on the PyQVM.

        Note that the state of the instance is reset on each call to ``execute``.

        :return: ``self`` to support method chaining.
        """
        if not isinstance(executable, Program):
            raise TypeError("`executable` argument must be a `Program`")

        self.program = executable
        self._memory_results = {}

        self.ram = {}
        if memory_map:
            self.ram.update(*memory_map)

        self.wf_simulator.reset()

        # grab the gate definitions for future use
        self._extract_defined_gates()

        self._memory_results = {}
        for _ in range(self.program.num_shots):
            self.wf_simulator.reset()
            self._execute_program()
            for name in self.ram.keys():
                self._memory_results.setdefault(name, list())
                self._memory_results[name].append(self.ram[name])

        self._memory_results = {k: np.asarray(v) for k, v in self._memory_results.items()}

        self._bitstrings = self._memory_results.get("ro")

        return self

    def get_result(self, execute_response: "PyQVM") -> QAMExecutionResult:
        """Return results from the PyQVM according to the common QAM API.

        Note that while the ``execute_response`` is not used, it's accepted in order to conform to that API; it's
        unused because the PyQVM, unlike other QAM's, is itself stateful.
        """
        if self.program is None:
            raise RuntimeError("No program loaded into PyQVM.")
        result_data = QVMResultData.from_memory_map(
            {key: RegisterData(matrix.tolist()) for key, matrix in self._memory_results.items()}
        )
        result_data = ResultData(result_data)
        data = ExecutionData(result_data=result_data, duration=None)
        return QAMExecutionResult(
            executable=self.program.copy(),
            data=data,
        )

    def read_memory(self, *, region_name: str) -> np.ndarray:
        """Read memory from the PyQVM."""
        if self._memory_results is None:
            raise ValueError("No memory results available.")
        return np.asarray(self._memory_results[region_name])

    def find_label(self, label: Union[Label, LabelPlaceholder]) -> int:
        """Iterate over the program and find a JumpTarget that has a Label matching the input label.

        :param label: Label object to search for in program
        :return: Program index where ``label`` is found
        """
        if self.program is None:
            raise RuntimeError("No program loaded into PyQVM.")
        for index, action in enumerate(self.program):
            if isinstance(action, JumpTarget):
                if label == action.label:
                    return index

        raise RuntimeError("Improper program - Jump Target not found in the input program!")

    def transition(self) -> bool:
        """Perform a QAM-like transition.

        This function assumes ``program`` and ``program_counter`` instance variables are set
        appropriately, and that the wavefunction simulator and classical memory ``ram`` instance
        variables are in the desired QAM input state.

        :return: whether the QAM should halt after this transition.
        """
        if self.program is None:
            raise RuntimeError("No program loaded into PyQVM.")
        instruction = self.program[self.program_counter]

        if isinstance(instruction, Gate):
            qubits = instruction.get_qubit_indices()
            if instruction.name in self.defined_gates:
                self.wf_simulator.do_gate_matrix(
                    matrix=self.defined_gates[instruction.name],
                    qubits=qubits,
                )
            else:
                self.wf_simulator.do_gate(gate=instruction)

            for noise_type, noise_prob in self.post_gate_noise_probabilities.items():
                self.wf_simulator.do_post_gate_noise(noise_type, noise_prob, qubits=qubits)

            self.program_counter += 1

        elif isinstance(instruction, Measurement):
            measured_val = self.wf_simulator.do_measurement(qubit=instruction.get_qubit_indices().pop())
            meas_reg: Optional[MemoryReference] = instruction.classical_reg
            if meas_reg is None:
                raise ValueError("Measurement instruction must have a classical register.")
            self.ram[meas_reg.name][meas_reg.offset] = measured_val
            self.program_counter += 1

        elif isinstance(instruction, Declare):
            if instruction.shared_region is not None:
                raise NotImplementedError("SHARING is not (yet) implemented.")

            self.ram[instruction.name] = list(
                np.zeros(instruction.memory_size, dtype=QUIL_TO_NUMPY_DTYPE[instruction.memory_type])
            )
            self.program_counter += 1

        elif isinstance(instruction, Pragma):
            # TODO: more stringent checks for what's being pragma'd and warnings
            self.program_counter += 1

        elif isinstance(instruction, Jump):
            # unconditional Jump; go directly to Label
            self.program_counter = self.find_label(instruction.target)

        elif isinstance(instruction, JumpTarget):
            # Label; pass straight over
            self.program_counter += 1

        elif isinstance(instruction, (JumpWhen, JumpUnless)):
            # JumpWhen/Unless; check classical reg
            jump_reg: Optional[MemoryReference] = instruction.condition
            if jump_reg is None:
                raise ValueError("JumpWhen/Unless instruction must have a classical register.")
            cond = self.ram[jump_reg.name][jump_reg.offset]
            if not isinstance(cond, (bool, np.bool_, np.int8, int)):
                raise ValueError(f"{type(instruction)} requires a data type of BIT; not {type(cond)}")
            dest_index = self.find_label(instruction.target)
            if isinstance(instruction, JumpWhen):
                jump_if_cond = True
            elif isinstance(instruction, JumpUnless):
                jump_if_cond = False
            else:
                raise TypeError(f"Invalid {type(instruction)}")

            if not (cond ^ jump_if_cond):
                # jumping: set prog counter to JumpTarget
                self.program_counter = dest_index
            else:
                # not jumping: hop over this instruction
                self.program_counter += 1

        elif isinstance(instruction, UnaryClassicalInstruction):
            # UnaryClassicalInstruction; set classical reg
            target = instruction.target
            old = self.ram[target.name][target.offset]
            if isinstance(instruction, ClassicalNeg):
                if not isinstance(old, (int, float, np.int_, np.float_)):
                    raise ValueError(f"NEG requires a data type of REAL or INTEGER; not {type(old)}")
                self.ram[target.name][target.offset] *= -1
            elif isinstance(instruction, ClassicalNot):
                if not isinstance(old, (bool, np.bool_)):
                    raise ValueError(f"NOT requires a data type of BIT; not {type(old)}")
                self.ram[target.name][target.offset] = not old
            else:
                raise TypeError("Invalid UnaryClassicalInstruction")

            self.program_counter += 1

        elif isinstance(instruction, (LogicalBinaryOp, ArithmeticBinaryOp, ClassicalMove)):
            left_ind = instruction.left
            left_val = self.ram[left_ind.name][left_ind.offset]
            if isinstance(instruction.right, MemoryReference):
                right_ind = instruction.right
                right_val = self.ram[right_ind.name][right_ind.offset]
            else:
                right_val = instruction.right

            if isinstance(instruction, ClassicalAnd):
                if not isinstance(left_val, int) or not isinstance(right_val, int):
                    raise ValueError("AND requires a data type of INTEGER; not {type(left_val)} and {type(right_val)}")
                new_val: Union[int, float] = left_val & right_val
            elif isinstance(instruction, ClassicalInclusiveOr):
                if not isinstance(left_val, int) or not isinstance(right_val, int):
                    raise ValueError("OR requires a data type of INTEGER; not {type(left_val)} and {type(right_val)}")
                new_val = left_val | right_val
            elif isinstance(instruction, ClassicalExclusiveOr):
                if not isinstance(left_val, int) or not isinstance(right_val, int):
                    raise ValueError("XOR requires a data type of INTEGER; not {type(left_val)} and {type(right_val)}")
                new_val = left_val ^ right_val
            elif isinstance(instruction, ClassicalAdd):
                new_val = left_val + right_val
            elif isinstance(instruction, ClassicalSub):
                new_val = left_val - right_val
            elif isinstance(instruction, ClassicalMul):
                new_val = left_val * right_val
            elif isinstance(instruction, ClassicalDiv):
                new_val = left_val / right_val
            elif isinstance(instruction, ClassicalMove):
                new_val = right_val
            else:
                raise ValueError(f"Unknown BinaryOp {type(instruction)}")
            self.ram[left_ind.name][left_ind.offset] = new_val
            self.program_counter += 1

        elif isinstance(instruction, ClassicalExchange):
            left_ind_ex = instruction.left
            right_ind_ex = instruction.right

            tmp = self.ram[left_ind_ex.name][left_ind_ex.offset]
            self.ram[left_ind_ex.name][left_ind_ex.offset] = self.ram[right_ind_ex.name][right_ind_ex.offset]
            self.ram[right_ind_ex.name][right_ind_ex.offset] = tmp
            self.program_counter += 1

        elif isinstance(instruction, Reset):
            self.wf_simulator.reset()
            self.program_counter += 1

        elif isinstance(instruction, ResetQubit):
            raise NotImplementedError("Need to implement in wf simulator")

        elif isinstance(instruction, Wait):
            self.program_counter += 1

        elif isinstance(instruction, Nop):
            # well that was easy
            self.program_counter += 1

        elif isinstance(instruction, DefGate):
            if instruction.parameters is not None and len(instruction.parameters) > 0:
                raise NotImplementedError("PyQVM does not support parameterized DEFGATEs")
            self.defined_gates[instruction.name] = instruction.matrix
            self.program_counter += 1

        elif isinstance(instruction, Halt):
            return True
        else:
            raise ValueError(f"Unsupported instruction type: {instruction}")

        # return HALTED (i.e. program_counter is end of program)
        if self.program is None:
            raise ValueError("No program loaded into PyQVM.")
        return self.program_counter == len(self.program)

    def _execute_program(self) -> "PyQVM":
        self.program_counter = 0

        if self.program is None:
            raise ValueError("No program loaded into PyQVM.")
        halted = len(self.program) == 0
        while not halted:
            halted = self.transition()

        return self

    def execute_once(self, program: Program) -> "PyQVM":
        """Execute one outer loop of a program on the PyQVM without re-initializing its state.

        Note that the PyQVM is stateful. Subsequent calls to :py:func:`execute_once` will not
        automatically reset the wavefunction or the classical RAM. If this is desired,
        consider starting your program with ``RESET``.

        :return: ``self`` to support method chaining.
        """
        self.program = program
        self._extract_defined_gates()
        return self._execute_program()



##############################################################################
# Copyright 2016-2019 Rigetti Computing
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################
"""Miscellaneous tools that are helpful for simulation."""

from collections.abc import Sequence
from typing import Union, cast

import numpy as np

from pyquil.experiment._setting import TensorProductState
from pyquil.paulis import PauliSum, PauliTerm
from pyquil.quil import Program
from pyquil.quilatom import Parameter
from pyquil.quilbase import Gate, Halt, _strip_modifiers
from pyquil.simulation.matrices import QUANTUM_GATES, STATES, SWAP


def all_bitstrings(n_bits: int) -> np.ndarray:
    """All bitstrings in lexicographical order as a 2d np.ndarray.

    This should be the same as ``np.array(list(itertools.product([0,1], repeat=n_bits)))``
    but faster.
    """
    n_bitstrings = 2**n_bits
    out = np.zeros(shape=(n_bitstrings, n_bits), dtype=np.int8)

    tf = np.array([False, True])
    for i in range(n_bits):
        # Lexicographical ordering gives a pattern of 1's
        # where runs of 1s of length 2**j are tiled 2**i times

        # i indexes from the *left*
        # j indexes from the *right*
        j = n_bits - i - 1

        out[np.tile(np.repeat(tf, 2**j), 2**i), i] = 1
    return out


def qubit_adjacent_lifted_gate(i: int, matrix: np.ndarray, n_qubits: int) -> np.ndarray:
    """Lift k-qubit gate on adjacent qubits from qubit i to complete Hilbert space of dimension 2 ** num_qubits.

    Ex: 1-qubit gate, lifts from qubit i
    Ex: 2-qubit gate, lifts from qubits (i+1, i)
    Ex: 3-qubit gate, lifts from qubits (i+2, i+1, i), operating in that order

    In general, this takes a k-qubit gate (2D matrix 2^k x 2^k) and lifts
    it to the complete Hilbert space of dim 2^num_qubits, as defined by
    the right-to-left tensor product (1) in arXiv:1608.03355.

    Developer note: Quil and the QVM like qubits to be ordered such that qubit 0 is on the right.
    Therefore, in ``qubit_adjacent_lifted_gate``, ``lifted_pauli``, and ``lifted_state_operator``,
    we build up the lifted matrix by performing the kronecker product from right to left.

    Note that while the qubits are addressed in decreasing order,
    starting with num_qubit - 1 on the left and ending with qubit 0 on the
    right (in a little-endian fashion), gates are still lifted to apply
    on qubits in increasing index (right-to-left) order.

    :param i: starting qubit to lift matrix from (incr. index order)
    :param matrix: the matrix to be lifted
    :param n_qubits: number of overall qubits present in space

    :return: matrix representation of operator acting on the
        complete Hilbert space of all num_qubits.
    """
    n_rows, n_cols = matrix.shape
    if n_rows != n_cols:
        raise ValueError("Matrix must be square")
    gate_size = np.log2(n_rows)
    if gate_size != int(gate_size):
        raise ValueError("Matrix must be 2^n by 2^n")
    gate_size = int(gate_size)

    # Outer-product to lift gate to complete Hilbert space

    # bottom: i qubits below target
    bottom_matrix = np.eye(2**i, dtype=np.complex128)
    # top: Nq - i (bottom) - gate_size (gate) qubits above target
    top_qubits = n_qubits - i - gate_size
    top_matrix = np.eye(2**top_qubits, dtype=np.complex128)

    return np.kron(top_matrix, np.kron(matrix, bottom_matrix))


def two_swap_helper(j: int, k: int, num_qubits: int, qubit_map: np.ndarray) -> tuple[np.ndarray, np.ndarray]:
    """Generate the permutation matrix that permutes two single-particle Hilbert spaces into adjacent positions.

    ALWAYS swaps j TO k. Recall that Hilbert spaces are ordered in decreasing
    qubit index order. Hence, j > k implies that j is to the left of k.

    End results:
        j == k: nothing happens
        j > k: Swap j right to k, until j at ind (k) and k at ind (k+1).
        j < k: Swap j left to k, until j at ind (k) and k at ind (k-1).

    Done in preparation for arbitrary 2-qubit gate application on ADJACENT
    qubits.

    :param j: starting qubit index
    :param k: ending qubit index
    :param num_qubits: number of qubits in Hilbert space
    :param qubit_map: current index mapping of qubits
    :return: tuple of swap matrix for the specified permutation,
             and the new qubit_map, after permutation is made
    """
    if not (0 <= j < num_qubits and 0 <= k < num_qubits):
        raise ValueError("Permutation SWAP index not valid")

    perm = np.eye(2**num_qubits, dtype=np.complex128)
    new_qubit_map = np.copy(qubit_map)

    if j == k:
        # nothing happens
        return perm, new_qubit_map
    elif j > k:
        # swap j right to k, until j at ind (k) and k at ind (k+1)
        for i in range(j, k, -1):
            perm = qubit_adjacent_lifted_gate(i - 1, SWAP, num_qubits).dot(perm)
            new_qubit_map[i - 1], new_qubit_map[i] = new_qubit_map[i], new_qubit_map[i - 1]
    elif j < k:
        # swap j left to k, until j at ind (k) and k at ind (k-1)
        for i in range(j, k, 1):
            perm = qubit_adjacent_lifted_gate(i, SWAP, num_qubits).dot(perm)
            new_qubit_map[i], new_qubit_map[i + 1] = new_qubit_map[i + 1], new_qubit_map[i]

    return perm, new_qubit_map


def permutation_arbitrary(qubit_inds: Sequence[int], n_qubits: int) -> tuple[np.ndarray, np.ndarray, int]:
    """Generate the permutation matrix that permutes an arbitrary number of single-particle Hilbert spaces into adjacent positions.

    Transposes the qubit indices in the order they are passed to a
    contiguous region in the complete Hilbert space, in increasing
    qubit index order (preserving the order they are passed in).

    Gates are usually defined as `GATE 0 1 2`, with such an argument ordering
    dictating the layout of the matrix corresponding to GATE. If such an
    instruction is given, actual qubits (0, 1, 2) need to be swapped into the
    positions (2, 1, 0), because the lifting operation taking the 8 x 8 matrix
    of GATE is done in the little-endian (reverse) addressed qubit space.

    For example, suppose I have a Quil command CCNOT 20 15 10.
    The median of the qubit indices is 15 - hence, we permute qubits
    [20, 15, 10] into the final map [16, 15, 14] to minimize the number of
    swaps needed, and so we can directly operate with the final CCNOT, when
    lifted from indices [16, 15, 14] to the complete Hilbert space.

    Notes: assumes qubit indices are unique (assured in parent call).

    See documentation for further details and explanation.

    Done in preparation for arbitrary gate application on
    adjacent qubits.

    :param qubit_inds: Qubit indices in the order the gate is applied to.
    :param n_qubits: Number of qubits in system
    :return:
        perm - permutation matrix providing the desired qubit reordering
        qubit_arr - new indexing of qubits presented in left to right decreasing index order.
        start_i - starting index to lift gate from
    """  # noqa: E501
    # Begin construction of permutation
    perm = np.eye(2**n_qubits, dtype=np.complex128)

    # First, sort the list and find the median.
    sorted_inds = np.sort(qubit_inds)
    med_i = len(qubit_inds) // 2
    med = sorted_inds[med_i]

    # The starting position of all specified Hilbert spaces begins at
    # the qubit at (median - med_i)
    start = med - med_i
    # Array of final indices the arguments are mapped to, from
    # high index to low index, left to right ordering
    final_map = np.arange(start, start + len(qubit_inds))[::-1]
    start_i = final_map[-1]

    # Note that the lifting operation takes a k-qubit gate operating
    # on the qubits i+k-1, i+k-2, ... i (left to right).
    # two_swap_helper can be used to build the
    # permutation matrix by filling out the final map by sweeping over
    # the qubit_inds from left to right and back again, swapping qubits into
    # position. we loop over the qubit_inds until the final mapping matches
    # the argument.
    qubit_arr = np.arange(n_qubits)  # current qubit indexing

    made_it = False
    right = True
    while not made_it:
        array = range(len(qubit_inds)) if right else range(len(qubit_inds))[::-1]
        for i in array:
            pmod, qubit_arr = two_swap_helper(
                np.where(qubit_arr == qubit_inds[i])[0][0], final_map[i], n_qubits, qubit_arr
            )

            # update permutation matrix
            perm = pmod.dot(perm)
            if np.allclose(qubit_arr[final_map[-1] : final_map[0] + 1][::-1], qubit_inds):
                made_it = True
                break

        # for next iteration, go in opposite direction
        right = not right

    if not np.allclose(qubit_arr[final_map[-1] : final_map[0] + 1][::-1], qubit_inds):
        segment_values = qubit_arr[final_map[-1] : final_map[0] + 1][::-1]
        raise ValueError(
            f"Permutation matrix not constructed correctly. "
            f"Expected array segment {segment_values} to match qubit indices {qubit_inds}. "
            f"Segment values: {segment_values}, Expected values: {qubit_inds}"
        )

    return perm, qubit_arr[::-1], start_i


def lifted_gate_matrix(matrix: np.ndarray, qubit_inds: Sequence[int], n_qubits: int) -> np.ndarray:
    """Lift a unitary matrix to act on the specified qubits in a full ``n_qubits``-qubit Hilbert space.

    For 1-qubit gates, this is easy and can be achieved with appropriate kronning of identity
    matrices. For 2-qubit gates acting on adjacent qubit indices, it is also easy. However,
    for a multiqubit gate acting on non-adjactent qubit indices, we must first apply a permutation
    matrix to make the qubits adjacent and then apply the inverse permutation.

    :param matrix: A 2^k by 2^k matrix encoding an n-qubit operation, where ``k == len(qubit_inds)``
    :param qubit_inds: The qubit indices we wish the matrix to act on.
    :param n_qubits: The total number of qubits.
    :return: A 2^n by 2^n lifted version of the unitary matrix acting on the specified qubits.
    """
    n_rows, n_cols = matrix.shape
    if n_rows != n_cols:
        raise ValueError("Matrix must be square")
    gate_size = np.log2(n_rows)
    if gate_size != int(gate_size):
        raise ValueError("Matrix must be 2^n by 2^n")
    gate_size = int(gate_size)

    pi_permutation_matrix, final_map, start_i = permutation_arbitrary(qubit_inds, n_qubits)
    if start_i > 0:
        check = final_map[-gate_size - start_i : -start_i]
    else:
        # Python can't deal with `arr[:-0]`
        check = final_map[-gate_size - start_i :]
    np.testing.assert_allclose(check, qubit_inds)

    v_matrix = qubit_adjacent_lifted_gate(start_i, matrix, n_qubits)
    return np.dot(np.conj(pi_permutation_matrix.T), np.dot(v_matrix, pi_permutation_matrix))  # type: ignore


def lifted_gate(gate: Gate, n_qubits: int) -> np.ndarray:
    """Lift a pyquil :py:class:`Gate` in a full ``n_qubits``-qubit Hilbert space.

    This function looks up the matrix form of the gate and then dispatches to
    :py:func:`lifted_gate_matrix` with the target qubits.

    :param gate: A gate
    :param n_qubits: The total number of qubits.
    :return: A 2^n by 2^n lifted version of the gate acting on its specified qubits.
    """
    zero = np.eye(2)
    zero[1, 1] = 0
    one = np.eye(2)
    one[0, 0] = 0

    if any(isinstance(param, Parameter) for param in gate.params):
        raise TypeError("Cannot produce a matrix from a gate with non-constant parameters.")

    # The main source of complexity is in handling handling FORKED gates. Given
    # a gate with modifiers, such as `FORKED CONTROLLED FORKED RX(a,b,c,d) 0 1
    # 2 3`, we get a tree, as in
    #
    #               FORKED CONTROLLED FORKED RX(a,b,c,d) 0 1 2 3
    #                 /                                      \
    #    CONTROLLED FORKED RX(a,b) 1 2 3       CONTROLLED FORKED RX(c,d) 1 2 3
    #                |                                        |
    #         FORKED RX(a,b) 2 3                      FORKED RX(c,d) 2 3
    #          /          \                            /          \
    #      RX(a) 3      RX(b) 3                    RX(c) 3      RX(d) 3
    #
    # We recurse on this structure using _gate_matrix below.

    def _gate_matrix(gate: Gate) -> np.ndarray:
        if len(gate.modifiers) == 0:  # base case
            if len(gate.params) > 0:
                return QUANTUM_GATES[gate.name](*gate.params)  # type: ignore
            else:
                return QUANTUM_GATES[gate.name]  # type: ignore
        else:
            mod = gate.modifiers[0]
            if mod == "DAGGER":
                child = _strip_modifiers(gate, limit=1)
                return _gate_matrix(child).conj().T
            elif mod == "CONTROLLED":
                child = _strip_modifiers(gate, limit=1)
                matrix = _gate_matrix(child)
                return np.kron(zero, np.eye(*matrix.shape)) + np.kron(one, matrix)  # type: ignore
            elif mod == "FORKED":
                if len(gate.params) % 2 != 0:
                    raise ValueError("FORKED gates must have an even number of parameters.")
                p0, p1 = gate.params[: len(gate.params) // 2], gate.params[len(gate.params) // 2 :]
                child = _strip_modifiers(gate, limit=1)
                # handle the first half of the FORKED params
                child.params = p0
                mat0 = _gate_matrix(child)
                # handle the second half of the FORKED params
                child.params = p1
                mat1 = _gate_matrix(child)
                return np.kron(zero, mat0) + np.kron(one, mat1)
            else:
                raise TypeError(f"Unsupported gate modifier {mod}")

    matrix = _gate_matrix(gate)

    return lifted_gate_matrix(matrix=matrix, qubit_inds=gate.get_qubit_indices(), n_qubits=n_qubits)


def program_unitary(program: Program, n_qubits: int) -> np.ndarray:
    """Return the unitary of a pyQuil program.

    :param program: A program consisting only of :py:class:`Gate`.:
    :return: a unitary corresponding to the composition of the program's gates.
    """
    umat: np.ndarray = np.eye(2**n_qubits)
    for instruction in program:
        if isinstance(instruction, Gate):
            unitary = lifted_gate(gate=instruction, n_qubits=n_qubits)
            umat = unitary.dot(umat)
        elif isinstance(instruction, Halt):
            pass
        else:
            raise ValueError(
                "Can only compute program unitary for programs composed of `Gate`s. "
                f"Found unsupported instruction: {instruction}"
            )
    return umat


def lifted_pauli(pauli_sum: Union[PauliSum, PauliTerm], qubits: list[int]) -> np.ndarray:
    """Return a matrix corresponding to the tensor representation of the given PauliSum and qubits.

    Useful for generating the full Hamiltonian after a particular fermion to
    pauli transformation. For example:

    Converting a PauliSum X0Y1 + Y1X0 into the matrix

    .. code-block:: python

       [
           [0.0 + 0.0j, 0.0 + 0.0j, 0.0 + 0.0j, 0.0 - 2.0j],
           [0.0 + 0.0j, 0.0 + 0.0j, 0.0 + 0.0j, 0.0 + 0.0j],
           [0.0 + 0.0j, 0.0 + 0.0j, 0.0 + 0.0j, 0.0 + 0.0j],
           [0.0 + 2.0j, 0.0 + 0.0j, 0.0 + 0.0j, 0.0 + 0.0j],
       ]

    Developer note: Quil and the QVM like qubits to be ordered such that qubit 0 is on the right.
    Therefore, in ``qubit_adjacent_lifted_gate``, ``lifted_pauli``, and ``lifted_state_operator``,
    we build up the lifted matrix by performing the kronecker product from right to left.

    :param pauli_sum: Pauli representation of an operator
    :param qubits: list of qubits in the order they will be represented in the resultant matrix.
    :return: matrix representation of the pauli_sum operator
    """
    if isinstance(pauli_sum, PauliTerm):
        pauli_sum = PauliSum([pauli_sum])

    n_qubits = len(qubits)
    result_hilbert = np.zeros((2**n_qubits, 2**n_qubits), dtype=np.complex128)
    # left kronecker product corresponds to the correct basis ordering
    for term in pauli_sum.terms:
        term_hilbert = np.array([1])
        for qubit in qubits:
            term_hilbert = np.kron(QUANTUM_GATES[term[qubit]], term_hilbert)

        result_hilbert += term_hilbert * cast(complex, term.coefficient)

    return result_hilbert


def tensor_up(pauli_sum: Union[PauliSum, PauliTerm], qubits: list[int]) -> np.ndarray:
    """Return a matrix corresponding to the tensor representation of the given PauliSum and qubits.

    This is the same as :py:func:`lifted_pauli`. Nick R originally wrote this functionality
    and really likes the name ``tensor_up``. Who can blame him?

    :param pauli_sum: Pauli representation of an operator
    :param qubits: list of qubits in the order they will be represented in the resultant matrix.
    :return: matrix representation of the pauli_sum operator
    """
    return lifted_pauli(pauli_sum=pauli_sum, qubits=qubits)


def lifted_state_operator(state: TensorProductState, qubits: list[int]) -> np.ndarray:
    """Return a matrix corresponding to the tensored-up representation of the given state and qubits.

    Developer note: Quil and the QVM like qubits to be ordered such that qubit 0 is on the right.
    Therefore, in ``qubit_adjacent_lifted_gate``, ``lifted_pauli``, and ``lifted_state_operator``,
    we build up the lifted matrix by using the *left* kronecker product.

    :param state: The state
    :param qubits: list of qubits in the order they will be represented in the resultant matrix.
    """
    mat: np.ndarray = np.eye(1)
    for qubit in qubits:
        oneq_state = state[qubit]
        if oneq_state.qubit != qubit:
            raise ValueError(f"State qubit mismatch: {oneq_state.qubit} != {qubit}")
        state_vector = STATES[oneq_state.label][oneq_state.index][:, np.newaxis]
        state_matrix = state_vector @ state_vector.conj().T
        mat = np.kron(state_matrix, mat)
    return mat


def scale_out_phase(unitary1: np.ndarray, unitary2: np.ndarray) -> np.ndarray:
    """Return a matrix m equal to unitary1/θ where ɑ satisfies unitary2 = e^(iθ)·unitary1.

    :param unitary1: The unitary matrix from which the constant of
        proportionality should be scaled-out.
    :param unitary2: The reference matrix.

    :return: A matrix (same shape as the input matrices) with the
             constant of proportionality scaled-out.
    """
    # h/t quilc
    rescale_value = 1.0
    goodness_value = 0.0

    for j in range(unitary1.shape[0]):
        if np.abs(unitary1[j, 0]) > goodness_value:
            goodness_value = np.abs(unitary1[j, 0])
            rescale_value = unitary2[j, 0] / unitary1[j, 0]

    return rescale_value * unitary1


def unitary_equal(A: np.ndarray, B: np.ndarray) -> bool:
    """Check if two matrices are unitarily equal."""
    if A.shape != B.shape:
        raise ValueError("Matrices must have the same shape")
    dim = A.shape[0]
    return np.allclose(np.abs(np.trace(A.T.conjugate() @ B) / dim), 1.0)



##############################################################################
# Copyright 2016-2019 Rigetti Computing
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################
"""Definition of an ExperimentSetting.

Each ExperimentSetting corresponds to preparing a collection of qubits in a TensorProductState and measuring them in a
PauliTerm-defined basis.
"""

import logging
import re
from collections.abc import Generator, Iterable
from dataclasses import dataclass
from typing import Any, Optional, cast

from pyquil.paulis import PauliTerm, sI

log = logging.getLogger(__name__)


@dataclass(frozen=True)
class _OneQState:
    """A description of a named one-qubit quantum state.

    This can be used to generate pre-rotations for quantum process tomography. For example,
    X0_14 will generate the +1 eigenstate of the X operator on qubit 14. X1_14 will generate the
    -1 eigenstate. SIC0_14 will generate the 0th SIC-basis state on qubit 14.
    """

    label: str
    index: int
    qubit: int

    def __str__(self) -> str:
        return f"{self.label}{self.index}_{self.qubit}"

    @classmethod
    def from_str(cls, s: str) -> "_OneQState":
        ma = re.match(r"\s*(\w+)(\d+)_(\d+)\s*", s)
        if ma is None:
            raise ValueError(f"Couldn't parse '{s}'")
        return _OneQState(label=ma.group(1), index=int(ma.group(2)), qubit=int(ma.group(3)))


@dataclass(frozen=True)
class TensorProductState:
    """A description of a multi-qubit quantum state that is a tensor product of many _OneQStates states."""

    states: list[_OneQState]

    def __init__(self, states: Optional[Iterable[_OneQState]] = None):
        if states is None:
            states = []
        object.__setattr__(self, "states", list(states))

    def __mul__(self, other: "TensorProductState") -> "TensorProductState":
        return TensorProductState(self.states + other.states)

    def __str__(self) -> str:
        return " * ".join(str(s) for s in self.states)

    def __repr__(self) -> str:
        return f"TensorProductState[{self}]"

    def __getitem__(self, qubit: int) -> _OneQState:
        """Return the _OneQState at the given qubit."""
        for oneq_state in self.states:
            if oneq_state.qubit == qubit:
                return oneq_state
        raise IndexError()

    def __iter__(self) -> Generator[_OneQState, None, None]:
        yield from self.states

    def __len__(self) -> int:
        return len(self.states)

    def states_as_set(self) -> frozenset[_OneQState]:
        return frozenset(self.states)

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, TensorProductState):
            return False

        return self.states_as_set() == other.states_as_set()

    def __hash__(self) -> int:
        return hash(self.states_as_set())

    @classmethod
    def from_str(cls, s: str) -> "TensorProductState":
        if s == "":
            return TensorProductState()
        return TensorProductState(list(_OneQState.from_str(x) for x in s.split("*")))


def SIC0(q: int) -> TensorProductState:
    return TensorProductState([_OneQState(label="SIC", index=0, qubit=q)])


def SIC1(q: int) -> TensorProductState:
    return TensorProductState([_OneQState(label="SIC", index=1, qubit=q)])


def SIC2(q: int) -> TensorProductState:
    return TensorProductState([_OneQState(label="SIC", index=2, qubit=q)])


def SIC3(q: int) -> TensorProductState:
    return TensorProductState([_OneQState(label="SIC", index=3, qubit=q)])


def plusX(q: int) -> TensorProductState:
    return TensorProductState([_OneQState(label="X", index=0, qubit=q)])


def minusX(q: int) -> TensorProductState:
    return TensorProductState([_OneQState(label="X", index=1, qubit=q)])


def plusY(q: int) -> TensorProductState:
    return TensorProductState([_OneQState(label="Y", index=0, qubit=q)])


def minusY(q: int) -> TensorProductState:
    return TensorProductState([_OneQState(label="Y", index=1, qubit=q)])


def plusZ(q: int) -> TensorProductState:
    return TensorProductState([_OneQState(label="Z", index=0, qubit=q)])


def minusZ(q: int) -> TensorProductState:
    return TensorProductState([_OneQState(label="Z", index=1, qubit=q)])


def zeros_state(qubits: Iterable[int]) -> TensorProductState:
    return TensorProductState([_OneQState(label="Z", index=0, qubit=q) for q in qubits])


@dataclass(frozen=True, init=False)
class ExperimentSetting:
    """Input and output settings for a tomography-like experiment.

    Many near-term quantum algorithms take the following form:

     - Start in a pauli state
     - Prepare some ansatz
     - Measure it w.r.t. pauli operators

    Where we typically use a large number of (start, measure) pairs but keep the ansatz preparation
    program consistent. This class represents the (start, measure) pairs. Typically a large
    number of these :py:class:`ExperimentSetting` objects will be created and grouped into
    an :py:class:`Experiment`.

    :ivar additional_expectations: A list of lists, where each inner list specifies a qubit subset
        to calculate the joint expectation value for. This attribute allows users to extract
        simultaneously measurable expectation values from a single experiment setting.
    """

    in_state: TensorProductState
    out_operator: PauliTerm
    additional_expectations: Optional[list[list[int]]] = None

    def __init__(
        self,
        in_state: TensorProductState,
        out_operator: PauliTerm,
        additional_expectations: Optional[list[list[int]]] = None,
    ):
        object.__setattr__(self, "in_state", in_state)
        object.__setattr__(self, "out_operator", out_operator)
        object.__setattr__(self, "additional_expectations", additional_expectations)

    def _in_operator(self) -> PauliTerm:
        # Backwards compat
        pt = sI()
        for oneq_state in self.in_state.states:
            if oneq_state.label not in ["X", "Y", "Z"]:
                raise ValueError(f"Can't shim {oneq_state.label} into a pauli term. Use in_state.")
            if oneq_state.index != 0:
                raise ValueError(f"Can't shim {oneq_state} into a pauli term. Use in_state.")

            new_pt = pt * PauliTerm(op=oneq_state.label, index=oneq_state.qubit)
            pt = cast(PauliTerm, new_pt)

        return pt

    def __str__(self) -> str:
        return f"{self.in_state}→{self.out_operator.compact_str()}"

    def __repr__(self) -> str:
        return f"ExperimentSetting[{self}]"

    def serializable(self) -> str:
        return str(self)

    @classmethod
    def from_str(cls, s: str) -> "ExperimentSetting":
        """Opposite of str(expt)."""
        instr, outstr = s.split("→")
        return ExperimentSetting(
            in_state=TensorProductState.from_str(instr),
            out_operator=PauliTerm.from_compact_str(outstr),
        )



"""A module containing utilities for working with Quil-T calibrations."""

from collections.abc import Sequence
from dataclasses import dataclass
from typing import Any, Optional, Union

import quil.expression as quil_expr
import quil.instructions as quil_rs
from quil.program import CalibrationSet

from pyquil.quilatom import (
    ExpressionDesignator,
    MemoryReference,
    QubitDesignator,
    _convert_to_py_expression,
)
from pyquil.quilbase import (
    AbstractInstruction,
    DefCalibration,
    DefMeasureCalibration,
    _convert_to_py_qubit,
    _convert_to_rs_instruction,
)


class CalibrationError(Exception):
    """Base class for calibration errors."""

    pass


class CalibrationDoesntMatch(CalibrationError):
    """Raised when a calibration doesn't match an instruction."""

    pass


@dataclass
class CalibrationMatch:
    """A match between a calibration definition and an instruction."""

    cal: Union[DefCalibration, DefMeasureCalibration]
    settings: dict[Union[QubitDesignator, ExpressionDesignator], Any]


def _convert_to_calibration_match(
    instruction: Union[quil_rs.Gate, quil_rs.Measurement],
    calibration: Optional[Union[quil_rs.Calibration, quil_rs.MeasureCalibrationDefinition]],
) -> Optional[CalibrationMatch]:
    if isinstance(instruction, quil_rs.Gate) and isinstance(calibration, quil_rs.Calibration):
        target_qubits = instruction.qubits
        target_values: Sequence[Union[quil_expr.Expression, MemoryReference]] = instruction.parameters
        parameter_qubits = calibration.qubits
        parameter_values: Sequence[Union[quil_expr.Expression, MemoryReference]] = calibration.parameters
        py_calibration: Union[DefCalibration, DefMeasureCalibration] = DefCalibration._from_rs_calibration(calibration)
    elif isinstance(instruction, quil_rs.Measurement) and isinstance(calibration, quil_rs.MeasureCalibrationDefinition):
        target_qubits = [instruction.qubit]
        target_values = (
            [] if not instruction.target else [MemoryReference._from_rs_memory_reference(instruction.target)]
        )
        parameter_qubits = [] if not calibration.qubit else [calibration.qubit]
        parameter_values = [MemoryReference._from_parameter_str(calibration.parameter)]
        py_calibration = DefMeasureCalibration._from_rs_measure_calibration_definition(calibration)
    else:
        return None

    settings: dict[Union[QubitDesignator, ExpressionDesignator], Union[QubitDesignator, ExpressionDesignator]] = {
        _convert_to_py_qubit(param): _convert_to_py_qubit(qubit)
        for param, qubit in zip(parameter_qubits, target_qubits)
        if isinstance(param, MemoryReference) or param.is_variable()
    }
    settings.update(
        {
            _convert_to_py_expression(param): _convert_to_py_expression(value)
            for param, value in zip(parameter_values, target_values)
            if isinstance(param, MemoryReference) or param.is_variable()
        }
    )

    return CalibrationMatch(py_calibration, settings)


def match_calibration(
    instr: AbstractInstruction, cal: Union[DefCalibration, DefMeasureCalibration]
) -> Optional[CalibrationMatch]:
    """Match a calibration definition to an instruction.

    On a successful match, return a (possibly empty) dictionary mapping calibration
    arguments and parameters to their values.

    On a failure, return None.
    """
    calibration = _convert_to_rs_instruction(cal)
    instruction = _convert_to_rs_instruction(instr)
    if calibration.is_calibration_definition() and instruction.is_gate():
        instruction = _convert_to_rs_instruction(instr)
        gate = instruction.to_gate()
        calibration_set = CalibrationSet([calibration.to_calibration_definition()], [])
        matched_calibration: Optional[Union[quil_rs.Calibration, quil_rs.MeasureCalibrationDefinition]] = (
            calibration_set.get_match_for_gate(gate)
        )
        return _convert_to_calibration_match(gate, matched_calibration)

    if calibration.is_measure_calibration_definition() and instruction.is_measurement():
        instruction = _convert_to_rs_instruction(instr)
        measurement = instruction.to_measurement()
        calibration_set = CalibrationSet([], [calibration.to_measure_calibration_definition()])
        matched_calibration = calibration_set.get_match_for_measurement(measurement)
        return _convert_to_calibration_match(measurement, matched_calibration)

    return None



##############################################################################
# Copyright 2016-2019 Rigetti Computing
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################
"""Convert a ``Program`` to a LaTeX quantum circuit diagram.

Note: this is a deprecated module: Import from pyquil.latex instead.
"""

from typing import Optional

from deprecated.classic import deprecated

from pyquil.latex._diagram import DiagramSettings
from pyquil.quil import Program


@deprecated(
    version="4.0",
    reason="This module has been moved -- please import it as 'from pyquil.latex import to_latex' going forward",
)
def to_latex(circuit: Program, settings: Optional[DiagramSettings] = None) -> str:
    """Produce a circuit diagram in LaTeX for a given pyQuil Program."""
    from pyquil.latex._main import to_latex

    return to_latex(circuit, settings)
